/*  File: zmapGLibUtils.c
 *  Author: Ed Griffiths (edgrif@sanger.ac.uk)
 *  Copyright (c) 2006-2017: Genome Research Ltd.
 *-------------------------------------------------------------------
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *-------------------------------------------------------------------
 * This file is part of the ZMap genome database package
 * originally written by:
 * 
 *      Ed Griffiths (Sanger Institute, UK) edgrif@sanger.ac.uk
 *        Roy Storey (Sanger Institute, UK) rds@sanger.ac.uk
 *   Malcolm Hinsley (Sanger Institute, UK) mh17@sanger.ac.uk
 *       Gemma Guest (Sanger Institute, UK) gb10@sanger.ac.uk
 *      Steve Miller (Sanger Institute, UK) sm23@sanger.ac.uk
 *  
 * Description: A collection of "add-ons" to GLib that seem useful.
 *
 * Exported functions: See ZMap/zmapGLibUtils.h
 *-------------------------------------------------------------------
 */

#include <ZMap/zmap.hpp>

#include <stdio.h>
#include <string.h>

#include <ZMap/zmapFeature.hpp>
#include <ZMap/zmapUtils.hpp>
#include <ZMap/zmapGLibUtils.hpp>




/* Holds a single quark set. */
typedef struct _ZMapQuarkSetStruct
{
  gint block_size ;
  GHashTable *g_quark_ht;
  gchar **g_quarks ;
  GQuark g_quark_seq_id ;
} ZMapQuarkSetStruct ;



typedef struct
{
  GQuark id;
}DatalistFirstIDStruct, *DatalistFirstID;



static inline GQuark g_quark_new(ZMapQuarkSet quark_set, gchar *string_arg) ;
static void printCB(gpointer data, gpointer user_data) ;
static gint caseCompareFunc(gconstpointer a, gconstpointer b) ;

static gboolean getNthHashElement(gpointer key, gpointer value, gpointer user_data) ;

static void hashCopyListCB(gpointer key, gpointer value, gpointer user_data) ;
static void hashPrintListCB(gpointer key, gpointer value, gpointer user_data_unused) ;
static void hashPrintTableCB(gpointer key, gpointer value, gpointer user_data_unused);
static void destroyList(gpointer data) ;
static void mergehashCB(gpointer key, gpointer value, gpointer user_data) ;

static void insert_after(GList *donor, GList *recipient) ;


/*
 *                    External routines.
 */




/*!
 *                Additions to String utilities
 */

/* This routine removes all occurences of the given char from the target string,
 * this is done inplace so the returned string will be shorter.
 *
 * This is not an efficient implementation, it would be done better by
 * keeping "to" and "from" pointers and simply moving sequentially through
 * the string skipping over the char to be removed and only moving the
 * other chars....one for a rainy day....
 *  */
char *zMap_g_remove_char(char *string_arg, char ch)
{
  char *cp ;

  zMapReturnValIfFailSafe((string_arg && *string_arg), string_arg) ;

  cp = string_arg ;
  while (*cp)
    {
      if (*cp == ch)
        {
          char *cq ;

          cq = cp ;
          while (*cq)
            {
              *cq = *(cq + 1) ;

              cq++ ;
            }
          *cq = '\0' ;
        }

      cp++ ;
    }

  return string_arg ;
}


/* This function is like strstr() but case independent. The code is modified
 * from g_ascii_strcasecmp() from glib/gstrfuncs.c. This seems like a no-brainer
 * to be added to glib. */
gchar *zMap_g_ascii_strstrcasecmp(const gchar *haystack, const gchar *needle)
{
  gchar *match_start = NULL, *needle_start = (gchar *)needle ;
  gint c1, c2;

#define ISUPPER(c)		((c) >= 'A' && (c) <= 'Z')
#define	TOLOWER(c)		(ISUPPER (c) ? (c) - 'A' + 'a' : (c))

  g_return_val_if_fail (haystack != NULL, NULL);
  g_return_val_if_fail (needle != NULL, NULL);

  while (*haystack && *needle)
    {
      c1 = (gint)(guchar) TOLOWER (*haystack);
      c2 = (gint)(guchar) TOLOWER (*needle);

      if (c1 == c2)
        {
          if (!match_start)
            match_start = (gchar *)haystack ;
          haystack++ ;
          needle++ ;
        }
      else
        {
          match_start = NULL ;
          needle = needle_start ;
          haystack++ ;
        }
    }

  if ((*needle))
    match_start = NULL ;

  return match_start ;
}




/*
 *                Additions to GList
 */


#ifdef UTILISE_SINGLE_FUNCTION
/* Go through a list in reverse calling the supplied GFunc.
 * This was lifted from the glib code of g_list. */

/* changed this slightly to be the g_list_foreach_directional
 * function, as it was working differently to g_list_foreach, by
 * the addition of the g_list_last.  g_list_foreach doesn't do a
 * g_list_first, but then it "shouldn't" need to as the all their
 * function return the first item of the list again. Here we need to
 * get to the end of the list to do the reverse! but the only place
 * it's getting called from at the moment is zmapWindowDrawFeatures.c
 * positionColumns, where the list is the item list of the
 * foocanvasgroup.  the group holds a pointer to the end of the list,
 * so we can use that instead and the other function below...
 * However, in the dna highlighting it's better not to go all the way
 * to the end of the list... Maybe we need 2 functions as it is nice
 * to have the g_list_last encapsulated in reverse function...
 */
void zMap_g_list_foreach_reverse(GList *glist, GFunc func, gpointer user_data)
{
  glist = g_list_last(glist) ;

  while (glist)
    {
      GList *prev = glist->prev ;

      (*func)(glist->data, user_data) ;
      glist = prev ;
    }

  return ;
}
#endif

void zMap_g_list_foreach_directional(GList   *glist,
                                     GFunc    func,
                                     gpointer user_data,
                                     ZMapGListDirection forwd)
{
  while (glist)
    {
      GList *next = (forwd == ZMAP_GLIST_FORWARD ? glist->next : glist->prev);
      (*func) (glist->data, user_data);
      glist = next;
    }
  return ;
}


/*! Just like g_list_foreach() except that the ZMapGFuncCond function can return
 * FALSE to stop the foreach loop from executing.
 *
 * Returns FALSE if ZMapGFuncCond returned FALSE, TRUE otherwise.
 *
 *  */
gboolean zMap_g_list_cond_foreach(GList *glist, ZMapGFuncCond func, gpointer user_data)
{
  gboolean status = TRUE ;

  while (glist)
    {
      GList *next = glist->next ;

      if (!((*func)(glist->data, user_data)))
        {
          status = FALSE ;
          break ;
        }

      glist = next ;
    }

  return status ;
}


/*!
 * Finds the given element data and moves it to the given position in the list if
 * that is possible. Note that list indices start at zero.
 *
 * Returns the list unaltered if the element could not be moved.
 *
 *  */
GList *zMap_g_list_move(GList *glist, gpointer user_data, guint new_index)
{
  GList *new_list = glist ;

  if (new_index < g_list_length(glist))
    {
      GList *list_element ;

      if ((list_element = g_list_find(glist, user_data)))
        {
          new_list = g_list_remove(new_list, user_data) ;

          new_list = g_list_insert(new_list, user_data, new_index) ;
        }
    }

  return new_list ;
}


GList *zMap_g_list_append_unique(GList *glist, gpointer data)
{
  GList *l,*last = glist;

  for(l = glist; l ; l = l->next)
    {
      if(l->data == data)	/* already there, nothing to do */
        return glist;
      last = l;
    }

  /* we reached the end or there was no glist: add to the end */
  last = g_list_append(glist,data);

  if(!glist)
    glist = last;

  return glist;
}


/* filter out a list of matching items */
GList *zMap_g_list_grep(GList **list_inout, gpointer data, GCompareFunc func)
{
  GList *matched = NULL;
  GList *loop_list, *tmp, *inout;

  /* zMapAssert(list_inout); */
  if (!list_inout)
    return matched ;

  loop_list = inout = *list_inout;

  while(loop_list)
    {
      if(!func(loop_list->data, data))
        {
          tmp = loop_list->next;
          inout = g_list_remove_link(inout, loop_list);
          matched = g_list_concat(matched, loop_list);
          loop_list = tmp;
        }
      else
        loop_list = loop_list->next;
    }

  /* basically in case we removed the first element */
  *list_inout = inout;

  return matched;
}


/* Inserts donor list _after_ the node in recipient given by "point".
 * To insert at the front set point to 0, to insert at the end set point
 * to be the length of the list.
 *
 * If copy_donor is TRUE a copy of the donor list is inserted into recipient
 * otherwise the actual donor list is inserted. Note that only a shallow copy
 * is done currently....so the copy list will point at the same data as the
 * original donor list.
 *
 * If recipient is NULL then donor (or a copy of donor) is returned,
 * if donor is NULL then recipient is returned.
 *
 * It is the callers responsibility to ensure that recipient and donor
 * are separate lists.
 *
 * Returns the new head of the recipient list or NULL if any of:
 *
 *             recipient && donor are NULL
 *             recipient == donor (only a trivial check is done)
 *             recipient or donor are NOT the head nodes of lists
 *             if !(0 <= point <= list_length)
 */
GList *zMap_g_list_insert_list_after(GList *recipient, GList *donor, int point, gboolean copy_donor)
{
  GList *complete = NULL ;
  int list_length ;

  if ((recipient || donor)
      && (recipient != donor)
      && (recipient == g_list_first(recipient))
      && (donor == g_list_first(donor))
      && (point >= 0 && point <= (list_length = g_list_length(recipient))))
    {
      if (copy_donor)
        donor = g_list_copy(donor) ;

      if (!recipient)
        {
          recipient = donor ;
        }
      else if (!donor)
        {
          ;						    /* nothing to do. */
        }
      else if (point == 0)
        {
          recipient = g_list_concat(donor, recipient) ;
        }
      else if (point == list_length)
        {
          recipient = g_list_concat(recipient, donor) ;
        }
      else
        {
          recipient = g_list_nth(recipient, point) ;
          insert_after(donor, recipient) ;
        }

      complete = g_list_first(recipient);
    }

  return complete ;
}


GList *zMap_g_list_lower(GList *move_list, int positions)
{
  GList *before;

  if(positions < 1)
    return move_list;

  if(move_list->prev)
    {
      for(before = move_list->prev; positions && before; positions--)
        before = before->prev;
    }
  else
    before = NULL;

  if(before)
    {
      before = g_list_remove_link(before, move_list);
      if(before->next)
        insert_after(move_list, before);
      else
        before = g_list_concat(before, move_list);
      /* zMapAssert(g_list_find(before, move_list->data));*/
    }

  return g_list_first(move_list);
}

GList *zMap_g_list_raise(GList *move_list, int positions)
{
  GList *before;

  if(positions < 1)
    return move_list;

  for(before = move_list; positions && before; positions--)
    before = before->next;

  if(!before)
    before = g_list_last(move_list);

  if(before)
    {
      before = g_list_remove_link(before, move_list);
      if(before->next)
        insert_after(move_list, before);
      else
        before = g_list_concat(before, move_list);
      /* zMapAssert(g_list_find(before, move_list->data));*/
    }

  return g_list_first(move_list);
}



/*! Takes a list and splits it into two at new_list_head. The original
 * list is truncated and ends with the element _before_ new_list_head,
 * the new list starts with new_list_head and contains all the elements
 * from there up until the end of the original list.
 *
 * Returns new_list_head or NULL if new_list_head is not in list OR
 * if new_list_head == list.
 *  */
GList *zMap_g_list_split(GList *glist, GList *new_list_head)
{
  GList *new_list = NULL ;

  if (new_list_head != glist && g_list_first(new_list_head) == glist)
    {
      (new_list_head->prev)->next = NULL ;
      new_list_head->prev = NULL ;
      new_list = new_list_head ;
    }

  return new_list ;
}


// merge lists giving list of unique items
GList *zMap_g_list_merge(GList *a, GList *b)
{
  for(;b;b = b->next)
    {
      if(!g_list_find(a,b->data))
        a = g_list_prepend(a,b->data);
    }
  return(a);
}





/*
 *  Functions for lists of quarks of which we have plenty in zmap.
 */


/*!
 * Prints out the contents of a list assuming that each element is a GQuark. We have
 * lots of these in zmap so this is useful.
 *
 * @param quark_list             A GList where each data pointer is a GQuark.
 * @param list_name              Optional list name to be printed first.
 * @param new_line               FALSE print list on one line, TRUE print each element
 *                               on a new line.
 * @return                       <nothing>
 */
void zMap_g_list_quark_print(GList *quark_list, const char *list_name, gboolean new_line)
{
  /* zMapAssert(quark_list) ;*/
  if (!quark_list)
    return ;

  if (list_name)
    printf("\"%s\":%s", list_name, (new_line ? "\n" : "\t")) ;

  g_list_foreach(quark_list, printCB, GINT_TO_POINTER(new_line)) ;

  printf("\n") ;

  return ;
}


/* Return a list of quarks delimited by the string given in delimiter,
 * if delimiter is NULL the string " ; " is used.
 * g_free() this string when finished. */
gchar *zMap_g_list_quark_to_string(GList *l, const char *delimiter)
{
  char *str = NULL ;

  if (l)
    {
      GString *quark_str ;

      if (!delimiter)
        delimiter = " ; " ;

      quark_str = g_string_new(g_quark_to_string(GPOINTER_TO_UINT(l->data))) ;
      l = l->next ;

      while(l)
        {
          g_string_append_printf(quark_str, "%s%s", delimiter,
                                 g_quark_to_string(GPOINTER_TO_UINT(l->data))) ;
          l = l->next ;
        }

      str = g_string_free(quark_str, FALSE) ;
    }

  return str ;
}


GList *zMap_g_list_find_quark(GList *glist, GQuark str_quark)
{
  GList *result = NULL ;

  /* zMapAssert(glist && str_quark) ; */
  if (!glist || !str_quark)
    return result ;

  result = g_list_find_custom(glist, GINT_TO_POINTER(str_quark), caseCompareFunc) ;

  return result ;
}


/* Removes the given list of quarks from quark_list, only removes the first
 * occurrence of a quark. */
GList *zMap_g_list_remove_quarks(GList *quark_target_list, GList *quarks)
{
  GList *quark_list = NULL ;

  if (quark_target_list)
    {
      quark_list = quark_target_list ;

      while(quarks)
        {
          quark_list = g_list_remove(quark_list, quarks->data) ;
          quarks = quarks->next ;
        }
    }

  return quark_list ;
}









/*!
 *                Additions to GHash
 */


/* MH17: this function is only ever called with nth == 0 */
/* Returns nth hash table element, elements start with the zero'th. */
gpointer zMap_g_hash_table_nth(GHashTable *hash_table, guint nth)
{
  gpointer entry = NULL ;

  /* zMapAssert(hash_table && nth >= 0) ; */
  if (!hash_table)
    return entry ;

  if (g_hash_table_size(hash_table) > nth)
    entry = g_hash_table_find(hash_table, getNthHashElement, &nth) ;

  return entry ;
}


/* Supported values for data_format_str are: "gquark", "pointer"  */
void zMap_g_hash_table_print(GHashTable *hash_table, const char *data_format_str)
{
  g_hash_table_foreach(hash_table, hashPrintTableCB, (void *)data_format_str) ;

  return ;
}

/*! A set of functions to handle a hash where each key is a GQuark and each entry is a
 * simple list that can be directly copied. Several places in our code need to
 * construct and look at his list hence these functions.
 *
 * Copy function could easily be extended to allow deep list copying with a callback.
 *
 * NOTE WELL: There is a destroy function for the hash values and this WILL get
 * called by g_hash_table for inserts/replaces. Below you will see we copy lists
 * because as soon as we reinsert the list the old copy will be deleted by the
 * delete callback.
 *
 *  */


/* ! Make the hash of lists table. */
GHashTable *zMap_g_hashlist_create(void)
{
  GHashTable *new_hashlist = NULL ;

  new_hashlist = g_hash_table_new_full(NULL, NULL, NULL, destroyList) ;

  return new_hashlist ;
}


/* ! Insert a key value into the keyed glist. */
void zMap_g_hashlist_insert(GHashTable *hashlist, GQuark key, gpointer value)
{
  GList *glist ;

  /* Slightly tricky coding, if we don't find a glist the append creates a new glist,
   * otherwise we append to the existing glist. Either way we then _always_ replace
   * the glist in the hash as glist start in theory could change. BUT note that
   * we must copy the glist before we insert it because otherwise it is erased
   * by g_hash_table calling the existing entries delete function. */
  glist = (GList *)g_hash_table_lookup(hashlist, GINT_TO_POINTER(key)) ;

  /* This makes this as bad as a g_datalist... Well nearly.   */
  if (!g_list_find(glist, value))
    {
      glist = g_list_copy(glist) ;

      glist = g_list_append(glist, value) ;

      g_hash_table_insert(hashlist, GINT_TO_POINTER(key), glist) ;
    }

  return ;
}


/* ! Insert a list of key values into the keyed list, if replace is TRUE then the list
 * replaces any existing list, if FALSE and a list already exists then the function simply
 * returns. */
void zMap_g_hashlist_insert_list(GHashTable *hashlist, GQuark key, GList *key_values, gboolean replace_flag)
{
  /* Slightly tricky coding, the existing copy of the list will automatically
   * be free'd by g_hash_table calling our registered hash delete function. */
  if (replace_flag || !(g_hash_table_lookup(hashlist, GINT_TO_POINTER(key))))
    g_hash_table_insert(hashlist, GINT_TO_POINTER(key), key_values) ;

  return ;
}


/*! Make a copy of the hash table of featureset names to styles lists. */
GHashTable *zMap_g_hashlist_copy(GHashTable *orig_hashlist)
{
  GHashTable *new_hashlist = NULL ;

  new_hashlist = zMap_g_hashlist_create() ;

  g_hash_table_foreach(orig_hashlist, hashCopyListCB, new_hashlist) ;

  return new_hashlist ;
}



/*! Merge two featureset/style hashes, any elements from "in" are merged into
 * in_out, in is left unaltered. */
void zMap_g_hashlist_merge(GHashTable *in_out, GHashTable *in)
{

  g_hash_table_foreach(in, mergehashCB, in_out) ;

  return ;
}



/*! Print out the featureset -> styles lists. */
void zMap_g_hashlist_print(GHashTable *hashlist)
{
  g_hash_table_foreach(hashlist, hashPrintListCB, NULL) ;

  return ;
}


void zMap_g_hashlist_destroy(GHashTable *hashlist)
{
  g_hash_table_destroy(hashlist) ;

  return ;
}


// replacement for functions not avail in old versions of glib
// return a list of key->value structs

static void get_key_value(gpointer key, gpointer value, gpointer user)
{
  ZMapGHashIter data;
  GList **glist = (GList **) user;

  data = g_new0(ZMapGHashIterStruct,1);
  data->key = key;
  data->value = value;
  *glist = g_list_prepend(*glist,data); // faster than append
}

void  zMap_g_hash_table_iter_init(GList **iter, GHashTable *h)
{
  *iter = NULL ;

  if (h)
    g_hash_table_foreach(h, get_key_value, iter) ;

  return ;
}


static void get_hash_key(gpointer key, gpointer value, gpointer user)
{
  GList **glist = (GList **) user;


  *glist = g_list_prepend(*glist,key); // faster than append
}

void  zMap_g_hash_table_get_keys(GList **iter, GHashTable *h)
{
  *iter = NULL;
  if(h)
    g_hash_table_foreach(h,get_hash_key,iter);

  return ;
}


static void get_hash_val(gpointer key, gpointer value, gpointer user)
{
  GList **glist = (GList **) user;


  *glist = g_list_prepend(*glist,value); // faster than append
}

void  zMap_g_hash_table_get_data(GList **iter, GHashTable *h)
{
  *iter = NULL;
  if(h)
    g_hash_table_foreach(h,get_hash_val,iter);
}


// must iter through all to free memory
gboolean zMap_g_hash_table_iter_next(GList **iter, gpointer *key, gpointer *value)
{
  gboolean result = FALSE ;
  ZMapGHashIter data ;
  GList *glist = *iter ;

  if(glist)
    {
      data = (ZMapGHashIter)(glist->data) ;
      *iter = g_list_remove_link(*iter,glist);
      g_list_free_1(glist);

      *key = data->key;
      *value = data->value;
      g_free(data);

      result = TRUE ;
    }

  return result ;
}

// in case we decide to not iter through the whole lot
// is there a generic list free function?
// yes: g_list_free()
//void zMap_g_hash_table_iter_free(GList **list) { }



/*
 *                Additions to GDatalist
 */

#if MH17_NOT_USED
G_LOCK_DEFINE_STATIC(datalist_first);

static void get_datalist_length(GQuark key, gpointer data, gpointer user_data)
{
  gint *length = (gint *)user_data;

  (*length)++;

  return ;
}

static void get_first_datalist_key(GQuark id, gpointer data, gpointer user_data)
{
  DatalistFirstID key = (DatalistFirstID)user_data;

  if(!key->id)
    key->id = id;

  return ;
}


gpointer zMap_g_datalist_first(GData **datalist)
{
  DatalistFirstIDStruct key = {0};
  gpointer data;
  G_LOCK(datalist_first);
  g_datalist_foreach(datalist, get_first_datalist_key, &key);
  G_UNLOCK(datalist_first);
  data = g_datalist_id_get_data(datalist, key.id);

  return data;
}

gint zMap_g_datalist_length(GData **datalist)
{
  gint length = 0;
  G_LOCK(datalist_first);
  g_datalist_foreach(datalist, get_datalist_length, &length);
  G_UNLOCK(datalist_first);
  return length;
}

#endif


/*
 *                Additions to GArray
 */


/* My new routine for returning a pointer to an element in the array, the array will be
 * expanded to include the requested element if necessary. If "clear" was not set when
 * the array was created then the element may contain random junk instead of zeros.
 *
 * NOTE, sadly user has to pass a pointer to a GArray point because the underlying
 * glib g_array_set_size () function returns an array pointer so we have to return
 * one too. You would imagine the struct would remain the same but you never know... */
gpointer zMap_g_array_element(GArray **array_inout, guint index)
{
  gpointer element_ptr = NULL ;
  GArray *an_array = *array_inout ;
  guint new_length, element_size ;

  new_length = index + 1 ;				    /* Note it's a zero-based index. */

  /* If we need to expand the an_array then bump it up quite a bit to avoid repeated reallocation. */
  if (new_length > an_array->len)
    {
      new_length *= 0.5 ;

      an_array = g_array_set_size(an_array, new_length) ;

      *array_inout = an_array ;				    /* Return new pointer if an_array resized. */
    }

  element_size = g_array_get_element_size(an_array) ;

  element_ptr = an_array->data + (element_size * index) ;

  return element_ptr ;
}







/*
 *                Additions to GString
 *
 * GString is a very good package but there are some more operations it could usefully do.
 *
 */


/*!
 * Substitute one string for another in a GString.
 *
 * You should note that this routine simply uses the existing GString erase/insert
 * functions and so may not be incredibly efficient. If the target string was not
 * found the GString remains unaltered and FALSE is returned.
 *
 * @param string                 A valid GString.
 * @param target                 The string to be replaced.
 * @param source                 The string to be inserted.
 * @return                       TRUE if a string was replaced, FALSE otherwise.
 *  */
gboolean zMap_g_string_replace(GString *string_arg, const char *target, const char *source)
{
  gboolean result = FALSE ;
  int source_len ;
  int target_len ;
  int target_pos ;
  char *template_ptr ;


  target_len = strlen(target) ;
  source_len = strlen(source) ;

  template_ptr = string_arg->str ;
  while ((template_ptr = strstr(template_ptr, target)))
    {
      result = TRUE ;

      target_pos = template_ptr - string_arg->str ;

      string_arg = g_string_erase(string_arg, target_pos, target_len) ;

      string_arg = g_string_insert(string_arg, target_pos, source) ;

      template_ptr = string_arg->str + target_pos + source_len ; /* Shouldn't go off the end. */
    }

  return result ;
}





/*!
 *                Additions to GQuark
 *
 * We need quarks that are allocated in sets that we can throw away as zmap can
 * have many thousands of quarks for each zmap displayed.
 *
 * The following routines create, manipulate and destroy a set of quarks.
 *
 */


/*!
 * Create the quark set.
 *
 * The block size allows control over the number of string
 * pointers allocated at one go, this may be a good optimisation where there are
 * large numbers of quarks to be allocated. If block size is 0 it defaults to 512
 * which is the glib default.
 * The function returns a new quark set which should only be destroyed with a call
 * to zMap_g_quark_destroy_set().
 *
 * @param block_size             0 for success, anything else for failure.
 * @return                       ZMapQuarkSet quark set.
 *  */
ZMapQuarkSet zMap_g_quark_create_set(guint block_size)
{
  ZMapQuarkSet quark_set = NULL ;

  quark_set = g_new0(ZMapQuarkSetStruct, 1) ;

  if (block_size == 0)
    block_size = 512 ;
  quark_set->block_size = block_size ;

  quark_set->g_quark_ht = g_hash_table_new(g_str_hash, g_str_equal) ;

  quark_set->g_quarks = NULL ;

  quark_set->g_quark_seq_id = 0 ;

  return quark_set ;
}

/*!
 * Look for a string in the quark set, if the string is not found no quark is
 * created and 0 is returned.
 *
 * @param quark_set              A valid quark set.
 * @param string                 The string to look for in the set.
 * @return                       GQuark for the string or 0 if string not found.
 *  */
GQuark zMap_g_quark_try_string(ZMapQuarkSet quark_set, gchar *string_arg)
{
  GQuark quark = 0 ;

  /* zMapAssert(quark_set && string_arg) ;*/
  if (!quark_set || !string_arg)
    return quark ;

  quark = GPOINTER_TO_UINT(g_hash_table_lookup(quark_set->g_quark_ht, string_arg)) ;

  return quark ;
}


/*!
 * Return a quark for a given string, if the string is not found, the string
 * is added to the set and its quark returned.
 *
 * @param quark_set              A valid quark set.
 * @param string                 The string to look for in the set.
 * @return                       GQuark for the string or 0 if string not found.
 *  */
GQuark zMap_g_quark_from_string(ZMapQuarkSet quark_set, gchar *string_arg)
{
  GQuark quark = 0 ;

  /* May be too draconian to insist on *string_arg...just remove if so. */
  /* zMapAssert(quark_set && string_arg && *string_arg) ;*/
  if (!quark_set || !string_arg || !*string_arg)
    return quark ;

  if (!(quark = (gulong) g_hash_table_lookup(quark_set->g_quark_ht, string_arg)))
    quark = g_quark_new(quark_set, g_strdup (string_arg));

  return quark;
}

/*!
 * Return the string for a given quark, if the quark cannot be found in the set
 * NULL is returned.
 *
 * @param quark_set              A valid quark set.
 * @param quark                  The quark for which the string should be returned.
 * @return                       A string or NULL.
 *  */
gchar *zMap_g_quark_to_string(ZMapQuarkSet quark_set, GQuark quark)
{
  gchar* result = NULL;

  /* zMapAssert(quark_set && quark) ; */
  if (!quark_set || !quark)
    return result ;

  if (quark > 0 && quark <= quark_set->g_quark_seq_id)
    result = quark_set->g_quarks[quark - 1] ;

  return result;
}


/*!
 * Destroy a quark_set, freeing all of its resources <B>including</B> the strings
 * it holds (i.e. you should make copies of any strings you need after the set
 * has been destroyed.
 *
 * @param quark_set              A valid quark set.
 * @return                       nothing
 *  */
void zMap_g_quark_destroy_set(ZMapQuarkSet quark_set)
{
  guint i ;

  g_hash_table_destroy(quark_set->g_quark_ht) ;

  /* Free the individual strings and then the string table itself. */
  for (i = 0 ; i < quark_set->g_quark_seq_id ; i++)
    {
      g_free(quark_set->g_quarks[i]) ;
    }
  g_free(quark_set->g_quarks) ;

  g_free(quark_set) ;

  return ;
}



/*
 *                 Internal functions.
 */


/* donor should already not be part of recipient */
static void insert_after(GList *donor, GList *recipient)
{
  /* zMapAssert(recipient && recipient->next && donor);*/
  if (!recipient || !recipient->next || !donor )
    return ;

  /*
   * This next bit of code is confusing so here is a desciption...
   *
   * recipient A -> <- B -> <- F
   * donor     C -> <- D -> <- E
   * point = 2
   *
   * recipient[B]->next[F]->prev = g_list_last(donor)[E]
   * recipient[B]->next[F]->prev[E]->next = recipient->next[F]
   * recipient[B]->next = donor[C]
   * donor[C]->prev = recipient[B]
   *
   * complete =  A -> <- B -> <- C -> <- D -> <- E -> <- F
   *                       |   |                   |   |
   * so we've changed B->next, C->prev,      E->next & F->prev
   */
  recipient->next->prev = g_list_last(donor);
  recipient->next->prev->next = recipient->next;
  recipient->next = donor;
  donor->prev = recipient;

  return ;
}



/* HOLDS: g_quark_global_lock */
static inline GQuark g_quark_new (ZMapQuarkSet quark_set, gchar *string_arg)
{
  GQuark quark;

  /* Allocate another block of quark string_arg pointers if needed. */
  if (quark_set->g_quark_seq_id % quark_set->block_size == 0)
    quark_set->g_quarks = g_renew(gchar*, quark_set->g_quarks,
				  quark_set->g_quark_seq_id + quark_set->block_size) ;

  quark_set->g_quarks[quark_set->g_quark_seq_id] = string_arg ;
  quark_set->g_quark_seq_id++ ;
  quark = quark_set->g_quark_seq_id ;

  g_hash_table_insert(quark_set->g_quark_ht, string_arg, GUINT_TO_POINTER(quark)) ;

  return quark ;
}



static void printCB(gpointer data, gpointer user_data)
{
  GQuark quark = GPOINTER_TO_INT(data) ;
  gboolean new_line = GPOINTER_TO_INT(user_data) ;

  printf("\"%s\"%s", g_quark_to_string(quark), (new_line ? "\n" : "\t")) ;

  return ;
}



/*
The function takes two gconstpointer arguments, the GList element's data and the given user data
*/

static gint caseCompareFunc(gconstpointer a, gconstpointer b)
{
  gint result = -1 ;
  GQuark list_quark = GPOINTER_TO_INT(a), str_quark = GPOINTER_TO_INT(b) ;

  result = g_ascii_strcasecmp(g_quark_to_string(list_quark), g_quark_to_string(str_quark)) ;

  return result ;
}





/* A hack really, this is GHRFunc() called from g_hash_table_find() which
 * will return TRUE on encountering the nth hash element thus stopping the find function and
 * returning the nth element of the hash.
 *
 * This function is NOT thread safe.
 *  */
static gboolean getNthHashElement(gpointer key, gpointer value, gpointer user_data)
{
  gboolean found_element = FALSE ;
  static int n = 0 ;
  int nth = *((int *)user_data) ;

  if (n == nth)
    {
      found_element = TRUE ;
      n = 0 ;
    }
  else
    n++ ;

  return found_element ;
}







/* A GHFunc() to copy the list attached to a hash entry. */
static void hashCopyListCB(gpointer key, gpointer value, gpointer user_data)
{
  GList *orig_list = (GList *)value ;
  GHashTable *new_hash = (GHashTable *)user_data ;
  GList *new_list = NULL ;

  new_list = g_list_copy(orig_list) ;

  g_hash_table_insert(new_hash, key, new_list) ;

  return ;
}



/* A GHFunc() to print the list attached to a hash entry. */
static void hashPrintListCB(gpointer key, gpointer value, gpointer user_data_unused)
{
  GList *styles_list = (GList *)value ;

  printf("\"%s\":  ", g_quark_to_string(GPOINTER_TO_INT(key))) ;

  zMap_g_list_quark_print(styles_list, "Styles", FALSE) ;

  return ;
}


/* A GHFunc() to print the quark attached to a hash entry. */
static void hashPrintTableCB(gpointer key, gpointer value, gpointer user_data)
{
  char *data_format_str = (char *)user_data ;
  char *value_str ;

  if (!data_format_str || g_ascii_strcasecmp(data_format_str, "pointer") == 0)
    {
      value_str = g_strdup_printf("%p", value) ;
    }
  else if (g_ascii_strcasecmp(data_format_str, "gquark") == 0)
    {
      value_str = g_strdup_printf("%s", g_quark_to_string(GPOINTER_TO_INT(value))) ;
    }
  else if (g_ascii_strcasecmp(data_format_str, "feature") == 0)
    {
      value_str = g_strdup_printf("'%s', '%s'", g_quark_to_string(((ZMapFeatureAny)value)->original_id),
                                            g_quark_to_string(((ZMapFeatureAny)value)->unique_id)) ;
    }
  else if (g_ascii_strcasecmp(data_format_str, "style") == 0)
    {
      value_str = g_strdup_printf("'%s', '%s'", g_quark_to_string(((ZMapFeatureTypeStyle)value)->original_id),
                                            g_quark_to_string(((ZMapFeatureTypeStyle)value)->unique_id)) ;
    }
  else
    {
      value_str = g_strdup("unknown data value") ;
    }

  printf("\"%s\": %s\n", g_quark_to_string(GPOINTER_TO_INT(key)), value_str) ;

  g_free(value_str) ;

  return ;
}



/* A GDestroyNotify() to free Glists held as the data in a hash. */
static void destroyList(gpointer data)
{
  GList *glist = (GList *)data ;

  if (glist)
    g_list_free(glist) ;

  return ;
}




static void mergehashCB(gpointer key, gpointer value, gpointer user_data)
{
  GHashTable *ghash = (GHashTable *)user_data ;

  /* If the feature set id is not in the hash then copy its list of styles and add it to the
   * target hash. */
  if (!(g_hash_table_lookup(ghash, key)))
    {
      GList *copy_list ;

      copy_list = g_list_copy((GList *)value) ;

      zMap_g_hashlist_insert_list(ghash, GPOINTER_TO_INT(key), copy_list, TRUE) ;
    }


  return ;
}







